
#include	"ptunnel.h"
#include	<pthread.h>
#include	<stdarg.h>
//#include	<sys/unistd.h>
#include	<sys/types.h>
//#include	<sys/socket.h>
//#include	<netinet/in.h>
//#include	<arpa/inet.h>
//#include	<netdb.h>
#include	<stdio.h>
#include	<stdlib.h>
#include	<string.h>
#include	<time.h>
#include	<signal.h>
#include	<errno.h>
//#include	<sys/time.h>
//#include	<net/ethernet.h>
#include    <windows.h>


WSADATA		wsd;

pcap_if_t *alldevs;
char errbuf[PCAP_ERRBUF_SIZE];



pthread_mutex_t		chain_lock,
					bw_lock,
					num_threads_lock;

bool				quit					= false,
					unprivileged			= false,
					pcap					= false;
FILE				*log_file				= 0;

int					tcp_port				= -1,
					tcp_listen_port			= -1,
					log_level				= 99,
					mode					= kMode_proxy,
					num_threads				= 0,
					max_threads				= kMax_threads;
char				*dest_proxy_addr		= 0,
					*dest_tcp_addr			= 0,
					*password				= 0,
					*pcap_device_addr		= 0;
const int			tcp_receive_buf_len		= kDefault_buf_size,
					icmp_receive_buf_len	= kDefault_buf_size + kIP_header_size + kICMP_header_size + sizeof(ping_tunnel_pkt_t),
					pcap_buf_size			= (kDefault_buf_size + kIP_header_size + kICMP_header_size + sizeof(ping_tunnel_pkt_t)+64)*64;
char				pcap_filter_program[]	= "icmp"; // && (icmp[icmptype] = icmp-echo || icmp[icmptype] = icmp-echoreply)";
proxy_desc_t		*chain					= 0;



int		main(int argc, char *argv[]) {
	int		i, opt;

	srand(time(0));
	log_file		= stdout;
	opt				= kOpt_undefined;
	mode			= kMode_proxy;
	for (i=1;i<argc;i++) {
		if (strcmp(argv[i], "-p") == 0) {
			mode	= kMode_forward;
			opt	= kOpt_set_proxy_addr;
		}
		else if (strcmp(argv[i], "-x") == 0)
			opt	= kOpt_set_password;
		else if (strcmp(argv[i], "-lp") == 0)
			opt	= kOpt_set_tcp_port;
		else if (strcmp(argv[i], "-da") == 0)
			opt	= kOpt_set_tcp_dest_addr;
		else if (strcmp(argv[i], "-dp") == 0)
			opt	= kOpt_set_tcp_dest_port;
		else if (strcmp(argv[i], "-v") == 0)
			opt	= kOpt_set_verbosity;
		else if (strcmp(argv[i], "-m") == 0)
			opt = kOpt_set_max_threads;
		else if (strcmp(argv[i], "-u") == 0)
			unprivileged	= !unprivileged;
		else if (strcmp(argv[i], "-c") == 0)
			opt	= kOpt_set_pcap_device;
		else {
			switch (opt) {
				case kOpt_set_proxy_addr:
					dest_proxy_addr			= argv[i];
					break;
				case kOpt_set_password:
					password				= argv[i];
					break;
				case kOpt_set_tcp_port:
					tcp_listen_port			= atoi(argv[i]);
					break;
				case kOpt_set_tcp_dest_addr:
					dest_tcp_addr			= argv[i];
					break;
				case kOpt_set_tcp_dest_port:
					tcp_port				= atoi(argv[i]);
					break;
				case kOpt_set_max_threads:
					max_threads	= atoi(argv[i]);
					if (max_threads <= 0)
						max_threads	= kMax_threads;
					break;
				case kOpt_set_verbosity:
					log_level		= atoi(argv[i]);
					break;
				case kOpt_set_pcap_device:
					pcap_device_addr= argv[i];
					pcap			= 1;
					break;
				case kOpt_undefined:
					usage(argv[0]);
					return 1;
			}
			opt	= kOpt_undefined;
		}
	}
	pt_log(kLog_info, "Starting ptunnel v %d.%.2d.\n", kMajor_version, kMinor_version);
	pt_log(kLog_info, "(c) 2004-2005 Daniel Stoedle, daniels@stud.cs.uit.no\n");
	pt_log(kLog_info, "%s.\n", (mode == kMode_forward ? "Relaying packets from incoming TCP streams" : "Forwarding incoming ping packets over TCP"));
	
//	signal(SIGPIPE, SIG_IGN);
	pthread_mutex_init(&chain_lock, 0);
	pthread_mutex_init(&bw_lock, 0);
	pthread_mutex_init(&num_threads_lock, 0);
	
	
	
	
	
	if(WSAStartup( MAKEWORD(2,2), &wsd) != NO_ERROR)
	{
	    pt_log(kLog_error, "Error at WSAStartup\n");
		return -1;
	}

	//	Check mode and validate arguments
	if (mode == kMode_forward) {
		if (!dest_proxy_addr || !dest_tcp_addr || !tcp_port || !tcp_listen_port) {
			printf("One of the options are missing or invalid.\n");
			usage(argv[0]);
			return -1;
		}
		pt_forwarder();
	}
	else
		pt_proxy(0);
	
	//	Clean up
	if (log_file != stdout)
		fclose(log_file);
	
	pt_log(kLog_info, "ptunnel is exiting.\n");
	return 0;
}


void		usage(char *exec_name) {
	printf("ptunnel v %d.%.2d.\n", kMajor_version, kMinor_version);
	printf("Usage: %s -p <addr> -lp <port> -da <dest_addr> -dp <dest_port> [-m max_threads] [-x password] [-v verbosity]\n", exec_name);
	printf("       %s [-m max_threads] [-x password] [-v verbosity] [-c <device>]\n", exec_name);
	printf("   -p: Set address of peer running packet forwarder. This causes\n");
	printf("       ptunnel to operate in forwarding mode - the absence of this\n");
	printf("       option causes ptunnel to operate in proxy mode.\n");
	printf("  -lp: Set TCP listening port (only used when operating in forward mode)\n");
	printf("  -da: Set remote proxy destination address\n");
	printf("  -dp: Set remote proxy destionation port\n");
	printf("   -m: Set maximum number of threads\n");
	printf("   -x: Set password (not implemented)\n");
	printf("   -u: Run proxy in unprivileged mode. This causes the proxy to forward\n");
	printf("       packets using standard echo requests, instead of crafting custom echo replies.\n");
	printf("       Unprivileged mode will only work on some systems, and is in general less reliable\n");
	printf("       than running in privileged mode.\n");
	printf("   -v: Verbosity level (-1 to 4, where -1 is no output, and 4 is all output)\n");
	printf("   -c: Enable libpcap on the given device.\n\n");
	printf("Starting the proxy (needs to run as root):\n");
	printf(" [root #] %s\n", exec_name);
	printf("Starting a client (also needs root):\n");
	printf(" [root #] %s -p proxy.pingtunnel.com -lp 8000 -da login.domain.com -dp 22 -c eth0\n", exec_name);
	printf("And then using the tunnel to ssh to login.domain.com:\n");
	printf(" [user $] ssh -p 8000 localhost\n");
	printf("And that's it. Enjoy your tunnel!\n\n");
}

BOOL find_interface(char *name, u_int *addr)
{
pcap_if_t *alldevs;
pcap_if_t *d;
pcap_addr *a;

int i=0;
char errbuf[PCAP_ERRBUF_SIZE];


	/* Retrieve the device list */
	if (pcap_findalldevs_ex(PCAP_SRC_IF_STRING, NULL, &alldevs, errbuf) == -1)
	{
		fprintf(stderr,"Error in pcap_findalldevs: %s\n", errbuf);
		exit(1);
	}
	
	/* Print the list */
	for(d=alldevs; d; d=d->next)
	{
		for(a=d->addresses; a; a=a->next)
		{
			if(((sockaddr_in*)(a->addr))->sin_addr.S_un.S_addr == *addr)
			{
				strcpy(name, d->name);
				pcap_freealldevs(alldevs);
				return TRUE;
			}

		}

	}
	pcap_freealldevs(alldevs);
	return FALSE;
}



//	pt_forwarder:
//	Sets up a listening TCP socket, and forwards incoming connections
//	over ping packets.

void		pt_forwarder(void) {

	SOCKET					server_sock, new_sock;
	fd_set				set;
	struct timeval		time;
	struct sockaddr_in	addr, dest_addr;
	socklen_t			addr_len;
	pthread_t			pid;
	int					sock;
	uint32_t			dst_ip, dst_port, proxy_ip;
	struct hostent		*host_ent;
	uint16_t			rand_id;
	
	BOOL				yes = TRUE;

	pt_log(kLog_debug, "Starting forwarder..\n");
	//	Open our listening socket
	sock = socket(AF_INET, SOCK_STREAM, IPPROTO_TCP);
	
	if (sock == INVALID_SOCKET) {
		pt_log(kLog_error, "Error at socket()\n");
	    return;
	}

	if (setsockopt(sock, SOL_SOCKET, SO_REUSEADDR,(char*) &yes, sizeof(yes)) == SOCKET_ERROR) {
		pt_log(kLog_error, "Failed to set SO_REUSEADDR option on listening socket: %s\n", strerror(errno));
		close(sock);
		return;
	}
	addr.sin_family	= AF_INET;
	addr.sin_port = htons((u_short)tcp_listen_port);
	addr.sin_addr.s_addr = INADDR_ANY;
	memset(&(addr.sin_zero), 0, sizeof(addr.sin_zero));
	if (bind(sock, (struct sockaddr*)&addr, sizeof(struct sockaddr)) == -1) {
		pt_log(kLog_error, "Failed to bind listening socket: %s\n", strerror(errno));
		close(sock);
		return;
	}
	server_sock	= sock;
	//	Look up IP of final destination (ie, where the proxy should send the packets)
	host_ent		= gethostbyname(dest_tcp_addr);
	if(host_ent == NULL)
	{
		pt_log(kLog_error, "cannot resolve address '%s', wsa error code %i", dest_tcp_addr, WSAGetLastError());
		return;
	}
	
	dst_ip			= *(long*)host_ent->h_addr_list[0];
	dst_port		= htonl(tcp_port);
	//	Look up proxy's IP
	host_ent		= gethostbyname(dest_proxy_addr);
	proxy_ip		= *(long*)host_ent->h_addr_list[0];
	//	Fill out address structure
	memset(&dest_addr, 0, sizeof(struct sockaddr_in));
	dest_addr.sin_family			= AF_INET;
	dest_addr.sin_port				= 0;
	dest_addr.sin_addr.s_addr		= proxy_ip;
	pt_log(kLog_verbose, "Proxy IP address: %s\n", inet_ntoa(*((struct in_addr*)&proxy_ip)));
	
	listen(server_sock, 10);
	while (!quit) {
		FD_ZERO(&set);
		FD_SET(server_sock, &set);
		time.tv_sec		= 1;
		time.tv_usec	= 0;
		if (select(server_sock+1, &set, 0, 0, &time) > 0) {
			pt_log(kLog_info, "Incoming connection.\n");
			addr_len	= sizeof(struct sockaddr_in);
			new_sock	= accept(server_sock, (struct sockaddr*)&addr, &addr_len);
			if (new_sock < 0) {
				pt_log(kLog_error, "Accepting incoming connection failed.\n");
				continue;
			}
			pthread_mutex_lock(&num_threads_lock);
			if (num_threads <= 0) {
				pt_log(kLog_event, "No running proxy thread - starting it.\n", max_threads);
				if (pthread_create(&pid, 0, pt_proxy, 0) != 0) {
					pt_log(kLog_error, "Couldn't create thread! Dropping incoming connection.\n");
					close(new_sock);
				}
				addr	= dest_addr;
				rand_id	= (uint16_t)rand();
				create_and_insert_proxy_desc(rand_id, rand_id, new_sock, &addr, dst_ip, dst_port, kProxy_start, kUser_flag);
			}
			pthread_mutex_unlock(&num_threads_lock);
		}
	}
}


//	pt_proxy: This function does all the client and proxy stuff.
//
void*		pt_proxy(void *args) {

	fd_set				set;
	struct timeval		timeout;
	int					bytes;
	struct sockaddr_in	addr;
	socklen_t			addr_len;
	u_int				icmp_sock	= 0,
						max_sock	= 0;
	char				*buf;
	double				now;
	proxy_desc_t		*cur, *prev, *tmp;
	pcap_info_t			pc;
	struct hostent		*host_ent;
	
	pt_log(kLog_debug, "Starting ping proxy..\n");
	buf					= (char*) malloc(icmp_receive_buf_len);
	//	Start the thread, initialize protocol and ring states.
	pt_log(kLog_debug, "Creating icmp datagram socket..\n");
	if (unprivileged)
		icmp_sock		= socket(AF_INET, SOCK_DGRAM, IPPROTO_ICMP);
	else
		icmp_sock		= socket(AF_INET, SOCK_RAW, IPPROTO_ICMP);
	if (icmp_sock < 0) {
		pt_log(kLog_error, "Couldn't create raw socket: %s\n", strerror(errno));
		return 0;
	}
	max_sock			= icmp_sock+1;
	if (pcap) {
		if (!unprivileged) {
			pt_log(kLog_info, "Initializing pcap.\n");
			
			char pcap_device[512];
			host_ent = gethostbyname(pcap_device_addr);
			if(host_ent == NULL)
			{
				pt_log(kLog_error, "cannot resolve address '%s', wsa error code %i", dest_tcp_addr, WSAGetLastError());
				exit(1);
			}
			find_interface(pcap_device, (u_int*)host_ent->h_addr_list[0]);
			
			
			pc.pcap_err_buf		= (char*) malloc(PCAP_ERRBUF_SIZE);
			pc.pcap_data_buf	= (char*) malloc(pcap_buf_size);
			
			pc.pcap_desc		= pcap_open(pcap_device, pcap_buf_size, PCAP_OPENFLAG_PROMISCUOUS, 50, NULL, pc.pcap_err_buf);
			//pc.pcap_desc		= pcap_open_live(pcap_device, pcap_buf_size, 1, 50 , pc.pcap_err_buf); //ms
			if (pc.pcap_desc) {
				pcap_lookupnet(pcap_device, &pc.netp, &pc.netmask, pc.pcap_err_buf);
				pt_log(kLog_verbose, "Network: %s\n", inet_ntoa(*(struct in_addr*)&pc.netp));
				pt_log(kLog_verbose, "Netmask: %s\n", inet_ntoa(*(struct in_addr*)&pc.netmask));
				if (pcap_compile(pc.pcap_desc, &pc.fp, pcap_filter_program, 0, pc.netp) == -1) {
					pt_log(kLog_error, "Failed to compile pcap filter program.\n");
					pcap_close(pc.pcap_desc);
					pcap	= 0;
				}
				else if (pcap_setfilter(pc.pcap_desc, &pc.fp) == -1) {
					pt_log(kLog_error, "Failed to set pcap filter program.\n");
					pcap_close(pc.pcap_desc);
					pcap	= 0;
				}
			}
			else
			{
				pt_log(kLog_verbose, "pcap opening error");
				exit(1);
			}

			pc.pkt_q.head	= 0;
			pc.pkt_q.tail	= 0;
			pc.pkt_q.elems	= 0;
			//	Check if we have succeeded, and free stuff if not
			if (!pcap) {
				pt_log(kLog_error, "There were errors enabling pcap - pcap has been disabled.\n");
				free(pc.pcap_err_buf);
				free(pc.pcap_data_buf);
			}
		}
		else
			pt_log(kLog_info, "pcap disabled since we're running in unprivileged mode.\n");
	}
	
	//	Start forwarding :)
	pt_log(kLog_info, "Ping proxy is listening in %s mode.\n", (unprivileged ? "unprivileged" : "privileged"));
	while (!quit) {
		FD_ZERO(&set);
		FD_SET(icmp_sock, &set);
		max_sock		= icmp_sock+1;
		for (cur=chain;cur;cur=cur->next) {
			if (cur->sock) {
				FD_SET(cur->sock, &set);
				if (cur->sock >= max_sock)
					max_sock	= cur->sock+1;
			}
		}
		timeout.tv_sec		= 0;
		timeout.tv_usec		= 100;
		select(max_sock, &set, 0, 0, &timeout);	//	Don't care about return val, since we need to check for new states anyway..
		
		pthread_mutex_lock(&chain_lock);
		for (prev=0,cur=chain;cur && cur->sock;cur=tmp) {
			//	Client: If we're starting up, send a message to the remote end saying so,
			//	causing him to connect to our desired endpoint.
			if (cur->state == kProxy_start) {
				pt_log(kLog_verbose, "Sending proxy request.\n");
				cur->last_ack	= timeval_to_double();
				queue_packet(icmp_sock, cur->pkt_type, buf, 0, cur->id_no, cur->id_no, &cur->my_seq, cur->send_ring, &cur->send_idx, &cur->send_wait_ack, cur->dst_ip, cur->dst_port, cur->state | cur->type_flag, &cur->dest_addr, cur->next_remote_seq, &cur->send_first_ack, &cur->ping_seq);
				cur->state		= kProto_data;
			}
			if (cur->should_remove) {
				tmp	= cur->next;
				remove_proxy_desc(cur, prev);
				continue;
			}
			if (FD_ISSET(cur->sock, &set) && cur->send_wait_ack < kPing_window_size) {
				bytes		= recv(cur->sock, cur->buf, tcp_receive_buf_len, 0);
				if (bytes <= 0) {
					pt_log(kLog_info, "Connection closed or lost.\n");
					tmp	= cur->next;
					//	Send it twice, and let's hope one of them makes it through..
					queue_packet(icmp_sock, cur->pkt_type, 0, 0, cur->id_no, cur->icmp_id, &cur->my_seq, cur->send_ring, &cur->send_idx, &cur->send_wait_ack, 0, 0, kProto_close | cur->type_flag, &cur->dest_addr, cur->next_remote_seq, &cur->send_first_ack, &cur->ping_seq);
					queue_packet(icmp_sock, cur->pkt_type, 0, 0, cur->id_no, cur->icmp_id, &cur->my_seq, cur->send_ring, &cur->send_idx, &cur->send_wait_ack, 0, 0, kProto_close | cur->type_flag, &cur->dest_addr, cur->next_remote_seq, &cur->send_first_ack, &cur->ping_seq);
					remove_proxy_desc(cur, prev);
					//	No need to update prev
					continue;
				}
				queue_packet(icmp_sock, cur->pkt_type, cur->buf, bytes, cur->id_no, cur->icmp_id, &cur->my_seq, cur->send_ring, &cur->send_idx, &cur->send_wait_ack, 0, 0, cur->state | cur->type_flag, &cur->dest_addr, cur->next_remote_seq, &cur->send_first_ack, &cur->ping_seq);
			}
			prev	= cur;
			tmp		= cur->next;
		}
		pthread_mutex_unlock(&chain_lock);
		
		if (FD_ISSET(icmp_sock, &set)) {
			//	Fill out address structure
			//	Handle ping traffic
			addr_len	= sizeof(struct sockaddr);
			bytes		= recvfrom(icmp_sock, buf, icmp_receive_buf_len, 0, (struct sockaddr*)&addr, &addr_len);
			if (bytes < 0) {
				pt_log(kLog_error, "Error receiving packet on ICMP socket: %s\n", strerror(errno));
				break;
			}
			handle_packet(buf, bytes, 0, &addr);
		}
		
		pthread_mutex_lock(&chain_lock);
		now		= timeval_to_double();
		for (cur=chain;cur;cur=cur->next) {
			if (cur->recv_wait_send && cur->sock)
				send_packets(cur->recv_ring, &cur->recv_xfer_idx, &cur->recv_wait_send, &cur->sock);
			int	idx	= cur->send_first_ack;
			//	Check for any icmp packets requiring resend, and attempt to _only_ resend the first packet.
			if (cur->send_ring[idx].pkt && cur->send_ring[idx].last_resend+kResend_interval < now) {
				pt_log(kLog_debug, "Resending packet with seq-no %d.\n", cur->send_ring[idx].seq_no);
				cur->send_ring[idx].last_resend		= now;
				cur->send_ring[idx].pkt->seq		= htons(cur->ping_seq);
				cur->ping_seq++;
				cur->send_ring[idx].pkt->checksum	= 0;
				cur->send_ring[idx].pkt->checksum	= htons(calc_icmp_checksum((u_short*)cur->send_ring[idx].pkt, cur->send_ring[idx].pkt_len));
				//printf("ID: %d\n", htons(cur->send_ring[idx].pkt->identifier));
				sendto(icmp_sock, (char*) cur->send_ring[idx].pkt, cur->send_ring[idx].pkt_len, 0, (struct sockaddr*)&cur->dest_addr, sizeof(struct sockaddr));
			}
			if (cur->last_ack+1.0 < now && cur->send_wait_ack < kPing_window_size && cur->remote_ack_val+1 != cur->next_remote_seq) {
				cur->last_ack	= now;
				queue_packet(icmp_sock, cur->pkt_type, 0, 0, cur->id_no, cur->icmp_id, &cur->my_seq, cur->send_ring, &cur->send_idx, &cur->send_wait_ack, cur->dst_ip, cur->dst_port, kProto_ack | cur->type_flag, &cur->dest_addr, cur->next_remote_seq, &cur->send_first_ack, &cur->ping_seq);
			}
		}
		pthread_mutex_unlock(&chain_lock);
		if (pcap) {
			if (pcap_dispatch(pc.pcap_desc, 32, pcap_packet_handler, (u_char*)&pc.pkt_q) > 0) {
				pqueue_elem_t	*cur;
				pt_log(kLog_verbose, "pcap captured %d packets - handling them..\n", pc.pkt_q.elems);
				while (pc.pkt_q.head) {
					cur				= pc.pkt_q.head;
					addr.sin_addr.s_addr	= ((ip_packet_t*)(cur->data))->src_ip;
					handle_packet(cur->data, cur->bytes, 1, &addr);
					pc.pkt_q.head	= cur->next;
					free(cur);
					pc.pkt_q.elems--;
				}
				pc.pkt_q.tail		= 0;
				pc.pkt_q.head		= 0;
			}
		}
	}
	pt_log(kLog_debug, "Proxy exiting..\n");
	if (icmp_sock)
		close(icmp_sock);
	//	TODO: Clean up the other descs
	free(buf);
	pt_log(kLog_debug, "Ping proxy done\n");
	return 0;
}


//	pcap_packet_handler:
//	This is our callback function handling captured packets. We already know that the packets
//	are ICMP echo or echo-reply messages, so all we need to do is strip off the ethernet header
//	and append it to the queue descriptor (the refcon argument).
//	
//	Ok, the above isn't entirely correct (we can get other ICMP types as well). This function
//	also has problems when it captures packets on the loopback interfacee.

void		pcap_packet_handler(u_char *refcon, const struct pcap_pkthdr *hdr, const u_char* pkt) {
	pqueue_t		*q;
	pqueue_elem_t	*elem;
	ip_packet_t		*ip;
	
	//pt_log(kLog_verbose, "Packet handler: %d =? %d\n", hdr->caplen, hdr->len);
	q		= (pqueue_t*)refcon;
	elem	= (pqueue_elem_t*) malloc(sizeof(pqueue_elem_t)+hdr->caplen-sizeof(struct ether_header));
	memcpy(elem->data, pkt+sizeof(struct ether_header), hdr->caplen-sizeof(struct ether_header));
	ip		= (ip_packet_t*)elem->data;
	//	TODO: Add fragment support
	elem->bytes	= ntohs(ip->pkt_len);
	if (elem->bytes > hdr->caplen-sizeof(struct ether_header)) {
		pt_log(kLog_error, "Received fragmented packet - unable to reconstruct!\n");
		free(elem);
		return;
	}
	//elem->bytes	= hdr->caplen-sizeof(struct ether_header);
	elem->next	= 0;
	if (q->tail) {
		q->tail->next	= elem;
		q->tail			= elem;
	}
	else {
		q->head	= elem;
		q->tail	= elem;
	}
	q->elems++;
}



//	handle_proxy_packet:
//	Processes incoming ICMP packets for the proxy. The packet can come either from the
//	packet capture lib, or from the actual socket or both.
//	Input:	A buffer pointing at the start of an IP header, the buffer length and the proxy
//			descriptor chain.
//	Output:	The new chain-head pointer.

void		handle_packet(char *buf, int bytes, int is_pcap, struct sockaddr_in *addr) {
	ip_packet_t			*ip_pkt;
	icmp_echo_packet_t	*pkt;
	ping_tunnel_pkt_t	*pt_pkt;
	proxy_desc_t		*cur;
	uint32_t			type_flag, pkt_flag;
	
	if (bytes < sizeof(icmp_echo_packet_t)+sizeof(ping_tunnel_pkt_t))
		pt_log(kLog_verbose, "Skipping this packet - too short.\n");
	else {
		ip_pkt		= (ip_packet_t*)buf;
		pkt			= (icmp_echo_packet_t*)ip_pkt->data;
		pt_pkt		= (ping_tunnel_pkt_t*)pkt->data;
		if (ntohl(pt_pkt->magic) == kPing_tunnel_magic) {
			pt_pkt->state		= ntohl(pt_pkt->state);
			pkt->identifier		= ntohs(pkt->identifier);
			pt_pkt->id_no		= ntohs(pt_pkt->id_no);
			pt_pkt->seq_no		= ntohs(pt_pkt->seq_no);
			//	Find the relevant connection, if it exists
			pthread_mutex_lock(&chain_lock);
			for (cur=chain;cur;cur=cur->next) {
				if (cur->id_no == pt_pkt->id_no) //pkt->identifier)
					break;
			}
			pthread_mutex_unlock(&chain_lock);
			
			//	Handle the packet if it comes from "the other end." This is a bit tricky
			//	to get right, since we receive both our own and the other end's packets.
			//	Basically, a proxy will accept any packet from a user, regardless if it
			//	has a valid connection or not. A user will only accept the packet if there
			//	exists a connection to handle it.
			
			if (cur) {
				type_flag		= cur->type_flag;
				if (type_flag == kProxy_flag)
					cur->icmp_id	= pkt->identifier;
			}
			else
				type_flag		= kProxy_flag;
			
			pkt_flag		= pt_pkt->state & kFlag_mask;
			pt_pkt->state	&= ~kFlag_mask;
			pt_log(kLog_debug, "Recv: %d [%d] bytes [seq = %d] [type = %s] [ack = %d] [icmp = %d] [user = %s] [pcap = %d]\n", bytes, ntohl(pt_pkt->data_len), pt_pkt->seq_no, ((pt_pkt->state == kProto_data) ? "data" : ((pt_pkt->state == kProto_ack) ? "ack" : "start")), ntohl(pt_pkt->ack), pkt->type, (pkt_flag == kUser_flag ? "yes" : "no"), is_pcap);
			//printf("State: %d  Is_User: %d PCap: %d Seq: %d\n", pt_pkt->state, pkt_flag == kUser_flag, is_pcap, pt_pkt->seq_no);
			if ((pkt_flag == kUser_flag && type_flag == kProxy_flag) || (pkt_flag == kProxy_flag && type_flag == kUser_flag)) {
				pt_pkt->data_len	= ntohl(pt_pkt->data_len);
				pt_pkt->ack			= ntohl(pt_pkt->ack);
				if (pt_pkt->state == kProxy_start) {
					if (!cur && type_flag == kProxy_flag) {
						pt_log(kLog_info, "Starting new proxy session to %s:%d with ID %d\n", inet_ntoa(*(struct in_addr*)&pt_pkt->dst_ip), ntohl(pt_pkt->dst_port), pt_pkt->id_no); //pkt->identifier);
						cur			= create_and_insert_proxy_desc(pt_pkt->id_no, pkt->identifier, 0, addr, pt_pkt->dst_ip, ntohl(pt_pkt->dst_port), kProto_data, kProxy_flag);
					}
					else if (type_flag == kUser_flag) {
						pt_log(kLog_error, "Dropping proxy session request - we are not a proxy!\n");
						return;
					}
					else
						pt_log(kLog_error, "Dropping duplicate proxy session request.\n");
				}
				if (cur && cur->sock) {
					if (pt_pkt->state == kProto_data || pt_pkt->state == kProxy_start || pt_pkt->state == kProto_ack)
						handle_data(pkt, bytes, cur->recv_ring, &cur->recv_wait_send, &cur->recv_idx, &cur->next_remote_seq);
					else if (pt_pkt->state == kProto_close) {
						pt_log(kLog_info, "Received session close from remote peer.\n");
						cur->should_remove	= 1;
					}
					handle_ack((uint16_t)pt_pkt->ack, cur->send_ring, &cur->send_wait_ack, 0, cur->send_idx, &cur->send_first_ack, &cur->remote_ack_val, is_pcap);
				}
			}
			else if (!is_pcap) {
			//	pt_log(kLog_verbose, "OS autoresponse - dropping.\n");
				//	Ack-packets work a bit different for the proxy as opposed to the client. The client
				//	can simply rely on the remote peer's operating system for supplying a ping response.
				//	This doesn't work for proxies, since we (usually) send echo replies, which don't create
				//	an echo reply from the remote end (obviously). So for the proxy, we also use the ack
				//	field of the packet for acks. This is handled by handle_ack (below).
				
		//		if (cur && pkt_flag == type_flag)
		//			handle_ack(pt_pkt->seq_no, cur->send_ring, &cur->send_wait_ack, 1, 0, &cur->send_first_ack, &cur->remote_ack_val, is_pcap);
		//		else
		//			pt_log(kLog_verbose, "Dropping possible acknowledgement: Corresponding connection not found.\n");
			}
		}
		else
			pt_log(kLog_verbose, "Ignored incoming packet.\n");
	}
}



//	create_and_insert_proxy_desc: Creates a new proxy descriptor, linking it into
//	the descriptor chain. If the sock argument is 0, the function will establish
//	a TCP connection to the ip and port given by dst_ip, dst_port.

proxy_desc_t*		create_and_insert_proxy_desc(uint16_t id_no, uint16_t icmp_id, int sock, struct sockaddr_in *addr, uint32_t dst_ip, uint32_t dst_port, uint32_t init_state, uint32_t type) {
	proxy_desc_t	*cur;
	
	pt_log(kLog_debug, "Adding proxy desc to run loop. Type is %s. Will create socket: %s\n", (type == kUser_flag ? "user" : "proxy"), (sock ? "No" : "Yes"));
	cur						= (proxy_desc_t*) calloc(1, sizeof(proxy_desc_t));
	cur->id_no				= id_no;
	cur->dest_addr			= *addr;
	cur->dst_ip				= dst_ip;
	cur->dst_port			= dst_port;
	cur->icmp_id			= icmp_id;
	if (!sock) {
		cur->sock				= socket(AF_INET, SOCK_STREAM, 0);
		memset(addr, 0, sizeof(struct sockaddr_in));
		addr->sin_port			= htons((uint16_t)dst_port);
		addr->sin_addr.s_addr	= dst_ip;
		addr->sin_family		= AF_INET;
		//	Let's just assume success, shall we?
		connect(cur->sock, (struct sockaddr*)addr, sizeof(struct sockaddr_in));
	}
	else
		cur->sock			= sock;
	cur->state				= init_state;
	cur->type_flag			= type;
	if (cur->type_flag == kUser_flag)
		cur->pkt_type		= 8;
	else
		cur->pkt_type		= (unprivileged ? 8 : 0);
	cur->buf				= (char*) malloc(icmp_receive_buf_len);
	
	pthread_mutex_lock(&chain_lock);
	cur->next				= chain;
	chain					= cur;
	pthread_mutex_unlock(&chain_lock);
	return cur;
}


//	remove_proxy_desc: Removes the given proxy desc, freeing its resources.
//	Assumes that we hold the chain_lock.
//
void		remove_proxy_desc(proxy_desc_t *cur, proxy_desc_t *prev) {
	int		i;
	
	pt_log(kLog_debug, "Removing proxy descriptor.\n");
	if (cur->buf)
		free(cur->buf);
	cur->buf	= 0;
	for (i=0;i<kPing_window_size;i++) {
		if (cur->send_ring[i].pkt)
			free(cur->send_ring[i].pkt);
		cur->send_ring[i].pkt	= 0;
		if (cur->recv_ring[i])
			free(cur->recv_ring[i]);
		cur->recv_ring[i]		= 0;
	}
	close(cur->sock);
	cur->sock	= 0;
	if (prev)
		prev->next	= cur->next;
	else
		chain		= cur->next;
}


//	queue_packet:
//	Creates an ICMP packet descriptor, and sends it. The packet descriptor is added
//	to the given send ring, for potential resends later on.

int			queue_packet(int icmp_sock, uint8_t type, char *buf, int num_bytes, uint16_t id_no, uint16_t icmp_id, uint16_t *seq, icmp_desc_t ring[], int *insert_idx, int *await_send, uint32_t ip, uint32_t port, uint32_t state, struct sockaddr_in *dest_addr, uint16_t next_expected_seq, int *first_ack, uint16_t *ping_seq) {
	int					pkt_len	= sizeof(icmp_echo_packet_t)+sizeof(ping_tunnel_pkt_t)+num_bytes,
						err		= 0;
	icmp_echo_packet_t	*pkt	= 0;
	ping_tunnel_pkt_t	*pt_pkt	= 0;
	uint16_t			ack_val	= next_expected_seq-1;
	
	if (pkt_len % 2)
		pkt_len++;
	pkt						= (icmp_echo_packet_t*) malloc(pkt_len);
	pkt->type				= type;		//	ICMP Echo request
	pkt->code				= 0;		//	Must be zero (non-zero requires root)
	pkt->identifier			= htons(icmp_id);
	pkt->seq				= htons(*ping_seq);
	pkt->checksum			= 0;
	(*ping_seq)++;
	//	Add our information
	pt_pkt					= (ping_tunnel_pkt_t*)pkt->data;
	pt_pkt->magic			= htonl(kPing_tunnel_magic);
	pt_pkt->dst_ip			= ip;
	pt_pkt->dst_port		= port;
	pt_pkt->ack				= htonl(ack_val);
	pt_pkt->data_len		= htonl(num_bytes);
	pt_pkt->state			= htonl(state);
	pt_pkt->seq_no			= htons(*seq);
	pt_pkt->id_no			= htons(id_no);
	//	Copy user data
	memcpy(pt_pkt->data, buf, num_bytes);
	pkt->checksum			= htons(calc_icmp_checksum((u_short*)pkt, pkt_len));
	//	Send it!
	pt_log(kLog_debug, "Send: %d [%d] bytes [seq = %d] [type = %s] [ack = %d] [icmp = %d] [user = %s]\n", pkt_len, num_bytes, *seq, ((state & (~kFlag_mask)) == kProto_data ? "data" : ((state & (~kFlag_mask)) == kProto_ack ? "ack" : "start")), ack_val, type, ((state & kUser_flag) == kUser_flag ? "yes" : "no"));
	err						= sendto(icmp_sock, (char *)pkt, pkt_len, 0, (struct sockaddr*)dest_addr, sizeof(struct sockaddr));
	if (err < 0) {
		pt_log(kLog_error, "Failed to send ICMP packet: %s\n", strerror(errno));
		return -1;
	}
	else if (err != pkt_len)
		pt_log(kLog_error, "WARNING WARNING, didn't send entire packet\n");
	
	//	Update sequence no's and so on
	ring[*insert_idx].pkt			= pkt;
	ring[*insert_idx].pkt_len		= pkt_len;
	ring[*insert_idx].last_resend	= timeval_to_double();
	ring[*insert_idx].jitter		= ((double)(rand() & 0xFF)) / 128.0;
	ring[*insert_idx].seq_no		= *seq;
	ring[*insert_idx].icmp_id		= icmp_id;
	(*seq)++;
	if (!ring[*first_ack].pkt)
		*first_ack					= *insert_idx;
	(*await_send)++;
	(*insert_idx)++;
	if (*insert_idx >= kPing_window_size)
		*insert_idx	= 0;
	return 0;
}


//	send_packets:
//	Examines the passed-in ring, and forwards data in it over TCP.
//
void		send_packets(forward_desc_t *ring[], int *xfer_idx, int *await_send, u_int *sock) {
	forward_desc_t		*fwd_desc;
	int					bytes;
	
	while (*await_send > 0) {
		fwd_desc	= ring[*xfer_idx];
		if (!fwd_desc)	//	We haven't got this packet yet..
			break;
		if (fwd_desc->length > 0) {
			bytes		= send(*sock, &fwd_desc->data[fwd_desc->length - fwd_desc->remaining], fwd_desc->remaining, 0);
			if (bytes < 0) {
				printf("Weirdness.\n");
				//	TODO: send close stuff
				close(*sock);
				*sock	= 0;
				break;
			}
			fwd_desc->remaining	-= bytes;
		}
		if (!fwd_desc->remaining) {
			ring[*xfer_idx]	= 0;
			free(fwd_desc);
			(*xfer_idx)++;
			(*await_send)--;
			if (*xfer_idx >= kPing_window_size)
				*xfer_idx	= 0;
		}
		else
			break;
	}
}


//	handle_data:
//	Utility function for handling kProto_data packets, and place the data it contains
//	onto the passed-in receive ring.

void		handle_data(icmp_echo_packet_t *pkt, int total_len, forward_desc_t *ring[], int *await_send, int *insert_idx,  uint16_t *next_expected_seq) {
	ping_tunnel_pkt_t	*pt_pkt			= (ping_tunnel_pkt_t*)pkt->data;
	int					expected_len	= sizeof(ip_packet_t) + sizeof(icmp_echo_packet_t) + sizeof(ping_tunnel_pkt_t); // 20+8+28
	
	//	Place packet in the receive ring, in its proper place.
	//	This works as follows:
	//	-1. Packet == ack packet? Perform ack, and continue.
	//	0. seq_no < next_remote_seq, and absolute difference is bigger than w size => discard
	//	1. If seq_no == next_remote_seq, we have no problems; just put it in the ring.
	//	2. If seq_no > next_remote_seq + remaining window size, discard packet. Send resend request for missing packets.
	//	3. Else, put packet in the proper place in the ring (don't overwrite if one is already there), but don't increment next_remote_seq_no
	//	4. If packed was not discarded, process ack info in packet.
	
	expected_len	+= pt_pkt->data_len;
	expected_len	+= expected_len % 2;
	if (total_len < expected_len) {
		pt_log(kLog_error, "Packet not completely received: %d Should be: %d. For some reason, this error is fatal.\n", total_len, expected_len);
		pt_log(kLog_debug, "Data length: %d Total length: %d\n", pt_pkt->data_len, total_len);
		exit(0);
	}
	if (pt_pkt->seq_no == *next_expected_seq) {
		//	hmm, what happens if this test is true?
		if (!ring[*insert_idx]) {	//  && pt_pkt->state == kProto_data
		//	pt_log(kLog_debug, "Queing data packet: %d\n", pt_pkt->seq_no);
			ring[*insert_idx]	= create_fwd_desc(pt_pkt->seq_no, pt_pkt->data_len, pt_pkt->data);
			(*await_send)++;
			(*insert_idx)++;
		}
		else if (ring[*insert_idx])
			pt_log(kLog_debug, "Dup packet?\n");
		
		(*next_expected_seq)++;
		if (*insert_idx >= kPing_window_size)
			*insert_idx	= 0;
		//	Check if we have already received some of the next packets
		while (ring[*insert_idx]) {
			if (ring[*insert_idx]->seq_no == *next_expected_seq) {
				(*next_expected_seq)++;
				(*insert_idx)++;
				if (*insert_idx >= kPing_window_size)
					*insert_idx	= 0;
			}
			else
				break;
		}
	}
	else {
		int	r, s, d, pos;
		pos	= -1;		//	If pos ends up staying -1, packet is discarded.
		r	= *next_expected_seq;
		s	= pt_pkt->seq_no;
		d	= s - r;
		if (d < 0) {	//	This packet _may_ be old, or seq_no may have wrapped around
			d	= (s+0xFFFF) - r;
			if (d < kPing_window_size) {
				//	Counter has wrapped, so we should add this packet to the recv ring
				pos	= ((*insert_idx)+d) % kPing_window_size;
			}
		}
		else if (d < kPing_window_size)
			pos	= ((*insert_idx)+d) % kPing_window_size;
		
		if (pos != -1) {
			if (!ring[pos]) {
				pt_log(kLog_verbose, "Out of order. Expected: %d  Got: %d  Inserted: %d (cur = %d)\n", *next_expected_seq, pt_pkt->seq_no, pos, (*insert_idx));
				ring[pos]	= create_fwd_desc(pt_pkt->seq_no, pt_pkt->data_len, pt_pkt->data);
				(*await_send)++;
			}
		}
		//else
		//	pt_log(kLog_debug, "Packet discarded - outside receive window.\n");
	}
}


void		handle_ack(uint16_t seq_no, icmp_desc_t ring[], int *packets_awaiting_ack, int one_ack_only, int insert_idx, int *first_ack, uint16_t *remote_ack, int is_pcap) {
	int		i, j, k;
	ping_tunnel_pkt_t	*pt_pkt;
	
	if (*packets_awaiting_ack > 0) {
		if (one_ack_only) {
			for (i=0;i<kPing_window_size;i++) {
				if (ring[i].pkt && ring[i].seq_no == seq_no && !is_pcap) {
					pt_log(kLog_debug, "Received ack for only seq %d\n", seq_no);
					pt_pkt	= (ping_tunnel_pkt_t*)ring[i].pkt->data;
					*remote_ack	= (uint16_t)ntohl(pt_pkt->ack);	//	WARNING: We make the dangerous assumption here that packets arrive in order!
					free(ring[i].pkt);
					ring[i].pkt	= 0;
					(*packets_awaiting_ack)--;
					if (i == *first_ack) {
						for (j=1;j<kPing_window_size;j++) {
							k	= (i+j)%kPing_window_size;
							if (ring[k].pkt) {
								*first_ack	= k;
								break;
							}
							if (k == i)	//	we have looped through everything
								*first_ack	= insert_idx;
							j++;
						}
					}
					return;
				}
			}
		}
		else {
			int	i, can_ack = 0, count = 0;
			i	= insert_idx-1;
			if (i < 0)
				i	= kPing_window_size - 1;
			
			pt_log(kLog_debug, "Received ack-series starting at seq %d\n", seq_no);
			while (count < kPing_window_size) {
				if (!ring[i].pkt)
					break;
				
				if (ring[i].seq_no == seq_no)
					can_ack	= 1;
				else if (!can_ack)
					*first_ack	= i;
				
				if (can_ack) {
					free(ring[i].pkt);
					ring[i].pkt	= 0;
					(*packets_awaiting_ack)--;
				}
				i--;
				if (i < 0)
					i	= kPing_window_size - 1;
				count++;
			}
		}
	}
//	else
//		pt_log(kLog_verbose, "Dropping superfluous acknowledgement (no outstanding packets needing ack.)\n");
}



forward_desc_t*	create_fwd_desc(uint16_t seq_no, uint32_t data_len, char *data) {
	forward_desc_t	*fwd_desc;
	fwd_desc			= (forward_desc_t *) malloc(sizeof(forward_desc_t)+data_len);
	fwd_desc->seq_no	= seq_no;
	fwd_desc->length	= data_len;
	fwd_desc->remaining	= data_len;
	if (data_len > 0)
		memcpy(fwd_desc->data, data, data_len);
	return fwd_desc;
}


unsigned short	calc_icmp_checksum(unsigned short *data, int bytes) {
	unsigned long	sum;
	int				i;
	
	sum	= 0;
	for (i=0;i<bytes/2;i++) {
		sum	+= data[i];
	}
	sum	= (sum & 0xFFFF) + (sum >> 16);
	sum	= htons(0xFFFF - sum);
	return sum;
}


void	pt_log(int level, char *fmt, ...) {
	va_list	args;
	const char *header[]	= { "[err]: ",
								"[inf]: ",
								"[evt]: ",
								"[vbs]: ",
								"[dbg]: " };
	
	if (level <= log_level) {
		va_start(args, fmt);
		fprintf(log_file, "%s", header[level]), vfprintf(log_file, fmt, args);
		va_end(args);
		if (log_file != stdout)
			fflush(log_file);
	}
}


double			timeval_to_double(void) {
	double			result;

	result		= (double)GetTickCount() / 1000.0;
	return result;
}
